/*
 * test_moves_order_tools.h --- Tools for testing order of moves.
 *
 * Copyright (c) 2015 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of Natch.
 *
 * Natch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Natch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "common.h"

#include "test_moves_order_tools.h"
#include "main.h"
#include "pcpjtool.h"

void
find_pawns(
    const exploration_t* exploration, piece_index_t* white, piece_index_t* black)
{
    piece_index_t i;
    for (i = INDEX_ROI_BLANC; i <= INDEX_PION_NOIR_H; i++)
    {
        const piece_t* piece = &exploration->pieces[ i ];
        if (piece->typePiece == PION)
        {
            if (piece->camp == BLANC)
            {
                white[column(piece->caseInitiale)] = i;
            }
            else
            {
                black[column(piece->caseInitiale)] = i;
            }
        }
    }
}

void
find_move_after_capture(
    piece_index_t captured_piece,
    const exploration_t* exploration,
    piece_index_t* before_piece_index,
    unsigned int* before_move_index)
{
    while (captured_piece != PIECE_PAS_CAPTUREE)
    {
        const piece_t* p = &exploration->pieces[captured_piece];
        const piece_t* capturer = &exploration->pieces[p->pieceCapturante];
        int i;
        bool_t changed = FALSE;

        for (i = 0; i < capturer->nbDestination; i++)
        {
            if (capturer->destinations[i].capturedPiece == captured_piece)
            {
                if (i == capturer->nbDestination - 1)
                {
                    if (capturer->pieceCapturante == PIECE_PAS_CAPTUREE)
                    {
                        *before_piece_index = capturer->index;
                        *before_move_index = i;
                        return;
                    }
                    else
                    {
                        captured_piece = capturer->index;
                        changed = TRUE;
                        break;
                    }
                }
                else
                {
                    *before_piece_index = capturer->index;
                    *before_move_index = i + 1;
                    return;
                }
            }
        }
        if (!changed)
        {
            exit_with_error("find_move_after_capture");
        }
    }
}

/*--------------------------------------------------------------------------*/

static move_type_t
improve_constraint_type(piece_index_t piece_index,
                        unsigned int move_index,
                        move_type_t type,
                        const exploration_t* exploration)
{
    const piece_t* piece;
    destination_t destination;
    unsigned int nb_spare_moves;
    int deltaY;
    square_t from;

    if (type == e_First_And_Last_Move)
    {
        return e_First_And_Last_Move;
    }

    piece = &exploration->pieces[piece_index];
    destination = piece->destinations[move_index];

    if (destination.destination == CASE_ARRIVEE_QUELCONQUE)
    {
        return e_First_And_Last_Move;
    }

    if (destination.distance > 1)
    {
        return type;
    }

    nb_spare_moves = piece->camp == BLANC
        ? exploration->white_spare_moves
        : exploration->black_spare_moves;

    /* Knight cannot loose a tempo in one move. */
    if (nb_spare_moves < 1
        || (piece->typePiece == CAVALIER && nb_spare_moves < 2))
    {
        return e_First_And_Last_Move;
    }

    if (piece->typePiece != PION)
    {
        return type;
    }

    if (piece->casePromotion != CASE_PAS_DE_PROMOTION)
    {
        unsigned int i = 0;
        while (i <= move_index)
        {
            /* After promotion, it's like other pieces.  We can't improve. */
            if (piece->destinations[i].destination == piece->casePromotion)
                return type;
            i++;
        }
    }

    if (move_index == 0)
    {
        from = piece->caseInitiale;
    }
    else
    {
        from = piece->destinations[move_index-1].destination;
    }
    deltaY = row(from) - row(destination.destination);
    if (deltaY == 16 || deltaY == -16)
    {
        return e_First_And_Last_Move;
    }

    return type;
}

/*--------------------------------------------------------------------------*/

static void
print_one_move( FILE*			fd,
                piece_index_t		piece_index,
                unsigned int		move_index,
                move_type_t		type,
                const exploration_t*	exploration )
{
    square_t	from;
    square_t    to;
    piece_t	piece = exploration->pieces[ piece_index ];
    static char	move_type[] = { ' ', 'F', 'L', 'B' };
    char        fromS[3];
    char        toS[3];

    if ( move_index == 0 )
    {
        from = piece.caseInitiale ;
    }
    else
    {
        from = piece.destinations[move_index - 1].destination;
    }
    to = piece.destinations[move_index].destination;

    if (from == CASE_ARRIVEE_QUELCONQUE)
    {
        sprintf(fromS, "%s", "??");
    }
    else
    {
        sprintf(fromS, "%c%c", 'a' + column(from), '1' + (row(from) >> 4));
    }

    if (to == CASE_ARRIVEE_QUELCONQUE)
    {
        sprintf(toS, "%s", "??");
    }
    else
    {
        sprintf(toS, "%c%c", 'a' + column(to), '1' + (row(to) >> 4));
    }

    if (from == CASE_ARRIVEE_QUELCONQUE)
    {
        sprintf(fromS, "%s", "??");
    }
    else
    {
        sprintf(fromS, "%c%c", 'a' + column( from ), '1' + ( row( from ) >> 4 ));
    }

    fprintf( fd,
             "%c%s-%s",
             InitialePieces[ piece.typePiece ],
             fromS,
             toS);
    fprintf( fd, "(%d,%d", piece_index, move_index );
    if ( type != e_First_And_Last_Move )
    {
        fprintf( fd, ",%c)", move_type[ type ] );
    }
    else
    {
        fprintf( fd, ")  " );
    }

}

/*--------------------------------------------------------------------------*/

bool_t
add_one_order( struct moves_order_t*	moves_order,
               piece_index_t		before_piece_index,
               unsigned int		before_move_index,
               move_type_t		before_type,
               piece_index_t		after_piece_index,
               unsigned int		after_move_index,
               move_type_t		after_type,
               const exploration_t*	exploration,
               const char*		description )
{
    move_type_t bef_type = improve_constraint_type(
        before_piece_index, before_move_index, before_type, exploration);
    move_type_t aft_type = improve_constraint_type(
        after_piece_index, after_move_index, after_type, exploration);

    if ( MainPrintPositionLevel > 2 )
    {
        fprintf( MainFD, "Add constraint (%s) : ", description );
        print_one_move( MainFD,
                        before_piece_index,
                        before_move_index,
                        bef_type,
                        exploration );
        fprintf( MainFD, " -> " );
        print_one_move( MainFD,
                        after_piece_index,
                        after_move_index,
                        aft_type,
                        exploration );
    }
    bool_t	ok = add_move_order( moves_order,
                                     before_piece_index,
                                     before_move_index,
                                     bef_type,
                                     after_piece_index,
                                     after_move_index,
                                     aft_type );
    if ( MainPrintPositionLevel > 2 )
    {
        fprintf( MainFD, "\n" );
    }

    if (!ok && shouldPrintEliminated())
    {
        print_eliminated_for_move_order(
            moves_order,
            before_piece_index,
            before_move_index,
            bef_type,
            after_piece_index,
            after_move_index,
            aft_type,
            exploration,
            description);
    }

    return ok ;
}

/*--------------------------------------------------------------------------*/

void
print_eliminated_for_move_order(
    struct moves_order_t*	moves_order,
    piece_index_t		before_piece_index,
    unsigned int		before_move_index,
    move_type_t		        before_type,
    piece_index_t		after_piece_index,
    unsigned int		after_move_index,
    move_type_t		        after_type,
    const exploration_t*	exploration,
    const char*		        description)
{
    writeAssociation(
        "Position eliminated because of impossible order :" );
    print_moves_order( MainFD, moves_order, exploration );
    fprintf( MainFD, "Impossible order (%s) : ", description );
    print_one_move( MainFD,
                    before_piece_index,
                    before_move_index,
                    before_type,
                    exploration );
    fprintf( MainFD, " -> " );
    print_one_move( MainFD,
                    after_piece_index,
                    after_move_index,
                    after_type,
                    exploration );
    fprintf( MainFD, "\n" );
}
