/*
 * twin_squares.c --- twin squares package
 *
 * Copyright (c) 2001, 2002, 2009, 2015 by Pascal Wassong All Rights Reserved.
 *
 * This file is part of Natch.
 *
 * Natch is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Natch is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include	"twin_squares.h"
#include	"pcpj.h"
#include	"pcpjtool.h"

/*--------------------------------------------------------------------------*/

square_t		No_Twin_Square = 0xFE ;

/*--------------------------------------------------------------------------*/

/* #define	DEBUG_TWINS */
#ifdef DEBUG_TWINS
#include "pcpjtool.h"
static void print_twin_squares(const twin_squares_t* twin_squares);
#endif /* DEBUG_TWINS */

/*--------------------------------------------------------------------------*/

void
create_twin_squares(
    const piece_t*	pieces,
    unsigned int	nb_pieces,
    twin_squares_t*     twin_squares)
{
    unsigned int	nb_twins = 0 ;
    unsigned int	i ;
    one_twin_t*		twins = twin_squares->twins ;

    for ( i = 0 ; i < nb_pieces ; i++ )
    {
        square_delta_t	deltaX, deltaY ;
        column_t	depX, 	arrX   ;
        row_t		depY, 	arrY   ;
        piece_type_t	type   = pieces[ i ].typePiece ;
        colour_t	colour = pieces[ i ].camp ;
        /* caseArrivee may not be up to date, so we have to use the expression
         * below to obtain the destination square.
         */
        square_t destination =
            pieces[i].destinations[pieces[i].nbDestination - 1].destination;

        if ( pieces[ i ].castling != NO_CASTLING )
        {
            if ( type == ROI && pieces[ i ].nbDestination == 1
                 && ( ( colour == BLANC && NbCoupsBlancsRestants < 2 )
                      || ( colour == NOIR && NbCoupsNoirsRestants < 2 ) ) )
            {
                twins[ nb_twins ].first      = pieces[ i ].caseInitiale ;
                twins[ nb_twins ].second     = destination ;
                twins[ nb_twins ].direction  = e_horizontal;
                twins[ nb_twins ].type       = type ;
                twins[ nb_twins ].colour     = colour ;
                twins[ nb_twins ].origin     = pieces[i].caseInitiale;
                nb_twins++ ;
            }
            continue ;
        }

        if ( type == CAVALIER
             || pieces[ i ].pieceCapturante != PIECE_PAS_CAPTUREE )
        {
            continue ;
        }
        if ((type == PION && pieces[i].destinations[0].distance > 0)
            || (pieces[i].nbDestination == 1
                && pieces[i].destinations[0].distance == 1
                && ((colour == BLANC && NbCoupsBlancsRestants == 0)
                    || (colour == NOIR && NbCoupsNoirsRestants == 0))))
        {
            /* There are potentially twin squares. */
        }
        else
        {
            continue ;
        }

        /* Eliminates promoted pawns except if they are on the promotion square
         * and may never have moved.
         */
        if ( type == PION
             && pieces[ i ].casePromotion != CASE_PAS_DE_PROMOTION
             && ( pieces[ i ].nbDestination > 1
                  || ( colour == BLANC && NbCoupsBlancsRestants > 1 )
                  || ( colour == NOIR  && NbCoupsNoirsRestants  > 1 ) ) )
        {
            continue ;
        }


        depX = column( pieces[ i ].caseInitiale );
        depY = row( pieces[ i ].caseInitiale ) >> 4 ;
        arrX = column( destination );
        arrY = row( destination ) >> 4 ;
        deltaX = depX > arrX ? depX - arrX : arrX - depX ;
        deltaY = depY > arrY ? depY - arrY : arrY - depY ;

        if ( ( type != PION && type != ROI )
             || deltaY == 1
             || deltaX == deltaY
             || ( type == PION
                  && deltaX == 0
                  && pieces[ i ].nbDestination == 1
                  && ( ( colour == BLANC && NbResteACapturerNoires < 2 )
                       || ( colour == NOIR
                            && NbResteACapturerBlanches < 2 ) ) ) )
        {
            twins[ nb_twins ].first      = pieces[ i ].caseInitiale ;
            twins[ nb_twins ].second     = destination ;
            if (deltaX != deltaY)
            {
                twins[ nb_twins ].direction
                    = deltaX != 0 ? e_horizontal : e_vertical;
            }
            else
            {
                twins[ nb_twins ].direction
                    = (depX - arrX) == (depY - arrY) ? e_up_right : e_up_left;
            }
            twins[ nb_twins ].type     	 = type ;
            twins[ nb_twins ].colour     = colour ;
            twins[ nb_twins ].origin     = pieces[i].caseInitiale;
            nb_twins++ ;
        }
    }

    twin_squares->nb_twins = nb_twins ;

#ifdef DEBUG_TWINS
    writeAssociation("printing position ans twin_squares");
    print_twin_squares(twin_squares);
#endif /* DEBUG_TWINS */
}

void replace_twin_square(
    twin_squares_t* twin_squares,
    square_t origin,
    square_t first,
    square_t second)
{
    unsigned int i;
    for (i = 0; i < twin_squares->nb_twins; i++)
    {
        if (twin_squares->twins[i].origin == origin)
        {
            twin_squares->twins[i].first = first;
            twin_squares->twins[i].second = second;
            return;
        }
    }
}

void add_twin_square(
    twin_squares_t* twin_squares,
    square_t first,
    square_t second,
    twin_direction_t direction,
    piece_type_t type,
    colour_t colour,
    square_t origin)
{
    one_twin_t* twin = &twin_squares->twins[twin_squares->nb_twins];
    twin->first = first;
    twin->second = second;
    twin->direction = direction;
    twin->type = type;
    twin->colour = colour;
    twin->origin = origin;

    twin_squares->nb_twins++;
}

void remove_last_twin_square(twin_squares_t* twin_squares)
{
    twin_squares->nb_twins--;
}

square_t
find_twin_square(
    const twin_squares_t* twin_squares,
    square_t		square,
    twin_direction_t	direction,
    piece_type_t	type,
    square_t            origin)
{
    unsigned int i;
    const one_twin_t* twins = twin_squares->twins;
    for ( i = 0 ; i < twin_squares->nb_twins ; i++ )
    {
        twin_direction_t dir = twins[ i ].direction;
        if (direction == dir
            && twins[i].type != type
            && twins[i].origin != origin)
        {
            if ( square == twins[ i ].first )
            {
                return twins[ i ].second ;
            }
            else if ( square == twins[ i ].second )
            {
                return twins[ i ].first;
            }
        }
    }

    return No_Twin_Square ;
}

void
print_twin_squares(const twin_squares_t* twin_squares)
{
    unsigned int i;
    for ( i = 0 ; i < twin_squares->nb_twins ; i++ )
    {
        char* dir;
        switch (twin_squares->twins[ i ].direction)
        {
        case e_horizontal:
            dir = "horizontal";
            break;
        case e_vertical:
            dir = "vertical";
            break;
        case e_up_right:
            dir = "up-right";
            break;
        case e_up_left:
            dir = "up-left";
            break;
        default:
            dir = "unknown dir ???";
        }
        fprintf(MainFD,
                "Twins %d : %s-%s (%s, %s)\n",
                i + 1,
                square_to_string(twin_squares->twins[ i ].first),
                square_to_string(twin_squares->twins[ i ].second),
                square_to_string(twin_squares->twins[ i ].origin),
                dir);
    }
}
